# coding: utf-8


def wrap_list(value):
    return value if isinstance(value, (tuple, list)) else [value]


def flatten(items):
    return [item for level in items for item in wrap_list(level)]


def slice_first(item):
    if isinstance(item, (tuple, list)):
        return item[0]
    return item


def filter_duplicates(iterable):
    filtered_list, seen = [], set()
    for item in iterable:
        key = slice_first(item)

        if key in seen:
            continue

        seen.add(key)
        filtered_list.append(item)

    return filtered_list


def merge_dicts(*dict_args):
    """
    Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.

    https://stackoverflow.com/questions/38987/how-to-merge-two-python-dictionaries-in-a-single-expression
    """
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result



# coding: utf-8
import pandas as pd
from pypika import (MySQLQuery,
                    terms)
from pypika.terms import Interval

from fireant.database import Database


class Trunc(terms.Function):
    """
    Wrapper for a custom MySQL TRUNC function (installed via a custon FireAnt MySQL script
    """

    def __init__(self, field, date_format, alias=None):
        super(Trunc, self).__init__('dashmore.TRUNC', field, date_format, alias=alias)


class MySQLDatabase(Database):
    """
    MySQL client that uses the PyMySQL module.
    """
    # The pypika query class to use for constructing queries
    query_cls = MySQLQuery

    def __init__(self, database=None, host='localhost', port=3306,
                 user=None, password=None, charset='utf8mb4'):
        self.host = host
        self.port = port
        self.database = database
        self.user = user
        self.password = password
        self.charset = charset

    def connect(self):
        import pymysql

        return pymysql.connect(
            host=self.host, port=self.port, db=self.database,
            user=self.user, password=self.password,
            charset=self.charset,
            cursorclass=pymysql.cursors.Cursor,
        )

    def fetch(self, query):
        with self.connect().cursor() as cursor:
            cursor.execute(query)
            return cursor.fetchall()

    def fetch_dataframe(self, query):
        return pd.read_sql(query, self.connect())

    def trunc_date(self, field, interval):
        return Trunc(field, interval)

    def interval(self, **kwargs):
        return Interval(**kwargs)



class Paginator(object):
    def __init__(self, limit=0, offset=0, order=()):
        """
        Class for keeping track of pagination parameters

        :param limit: The number of rows that should be returned
        :param offset: The number of rows to offset the query by
        :param order: Collection of tuples in the format
                      (<metric/dimension key>, ``pypika.Order.desc`` or ``pypika.Order.asc``)
        """
        self.offset = offset
        self.limit = limit
        self.order = order

    def __str__(self):
        return 'offset: {offset} limit: {limit} order: {order}'.format(offset=self.offset,
                                                                       limit=self.limit,
                                                                       order=[(key, orderby.name)
                                                                              for key, orderby in self.order])



# coding: utf-8

from .base import (Transformer,
                   TransformationException)
from .datatables import (DataTablesRowIndexTransformer,
                         DataTablesColumnIndexTransformer)
from .datatables import (DataTablesRowIndexTransformer,
                         DataTablesColumnIndexTransformer,
                         CSVRowIndexTransformer,
                         CSVColumnIndexTransformer)
from .highcharts import (HighchartsLineTransformer,
                         HighchartsAreaTransformer,
                         HighchartsAreaPercentageTransformer,
                         HighchartsColumnTransformer,
                         HighchartsStackedColumnTransformer,
                         HighchartsBarTransformer,
                         HighchartsStackedBarTransformer,
                         HighchartsPieTransformer)
from .notebooks import (PandasRowIndexTransformer,
                        PandasColumnIndexTransformer,
                        MatplotlibLineChartTransformer,
                        MatplotlibBarChartTransformer)

ROW_INDEX_TABLE = 'row_index_table'
ROW_INDEX_CSV = 'row_index_csv'
COLUMN_INDEX_TABLE = 'column_index_table'
COLUMN_INDEX_CSV = 'column_index_csv'
LINE_CHART = 'line_chart'
BAR_CHART = 'bar_chart'
AREA_CHART = 'area_chart'
AREA_PERCENTAGE_CHART = 'area_percentage_chart'
COLUMN_CHART = 'column_chart'
STACKED_COLUMN_CHART = 'stacked_column_chart'
STACKED_BAR_CHART = 'stacked_bar_chart'
PIE_CHART = 'pie_chart'

BUNDLES = {
    'notebooks': {
        ROW_INDEX_TABLE: PandasRowIndexTransformer(),
        COLUMN_INDEX_TABLE: PandasColumnIndexTransformer(),
        LINE_CHART: MatplotlibLineChartTransformer(),
        BAR_CHART: MatplotlibBarChartTransformer(),
    },
    'highcharts': {
        LINE_CHART: HighchartsLineTransformer(),
        AREA_CHART: HighchartsAreaTransformer(),
        AREA_PERCENTAGE_CHART: HighchartsAreaPercentageTransformer(),
        COLUMN_CHART: HighchartsColumnTransformer(),
        STACKED_COLUMN_CHART: HighchartsStackedColumnTransformer(),
        BAR_CHART: HighchartsBarTransformer(),
        STACKED_BAR_CHART: HighchartsStackedBarTransformer(),
        PIE_CHART: HighchartsPieTransformer(),
    },
    'datatables': {
        ROW_INDEX_TABLE: DataTablesRowIndexTransformer(),
        COLUMN_INDEX_TABLE: DataTablesColumnIndexTransformer(),
        ROW_INDEX_CSV: CSVRowIndexTransformer(),
        COLUMN_INDEX_CSV: CSVColumnIndexTransformer(),
    },
}



# coding: utf-8
from .managers import SlicerException
from .filters import (
    BooleanFilter,
    ContainsFilter,
    EqualityFilter,
    ExcludesFilter,
    RangeFilter,
    WildcardFilter,
)
from .pagination import Paginator
from .schemas import (
    DimensionValue,
    EqualityOperator,
    Join,
    Metric,
    Slicer,
)
from .schemas import (
    DatetimeInterval,
    NumericInterval,
)
from .schemas import (
    BooleanDimension,
    CategoricalDimension,
    ContinuousDimension,
    DatetimeDimension,
    UniqueDimension,
)
